@@ -531,6 +531,7 @@ meta_core_get_menu_accelerator (MetaMenuOp           menu_op,
       name = META_KEYBINDING_TOGGLE_STICKY;
       break;
     case META_MENU_OP_ABOVE:
+    case META_MENU_OP_UNABOVE:
       name = META_KEYBINDING_TOGGLE_ABOVE;
       break;
     case META_MENU_OP_WORKSPACES:
@@ -630,7 +631,6 @@ meta_core_begin_grab_op (Display    *xdisplay,
                          MetaGrabOp  op,
                          gboolean    pointer_already_grabbed,
                          gboolean    frame_action,
-                         int         event_serial,
                          int         button,
                          gulong      modmask,
                          guint32     timestamp,
@@ -649,7 +649,6 @@ meta_core_begin_grab_op (Display    *xdisplay,
   return meta_display_begin_grab_op (display, screen, window,
                                      op, pointer_already_grabbed,
                                      frame_action,
-                                     event_serial,
                                      button, modmask,
                                      timestamp, root_x, root_y);
 }
